#!/usr/bin/perl

# "sendpages.pl" in the Forum News Gateway
# copyright 1994,1995 by Swarthmore College

# SendAccountCreationPage($ProtoVers,$Admin,$ExtraMessage)
# Send the page that lets a user or the administrator create an account.
# If $Admin=1, the administrator is creating the account; if 0, the user.
# If $ExtraMessage isn't "", include it in the page.

sub SendAccountCreationPage {
	local($ProtoVers,$Admin,$ExtraMessage) = @_;
	
	&SendProtoVers($ProtoVers,$BuiltInExpirationSeconds);
	
	if ($Admin) {
		&SendToClient("<HEAD><TITLE>Create newsreading accounts</TITLE></HEAD>\012<BODY>\012",
			"<H2>Create user newsreading accounts</H2>\012",
			"The user name and password can be almost anything.\012",
			"You must give the user's real name and complete e-mail address,\012",
			"and optionally, the name of the user's organization, so that articles\012",
			"the user posts have the proper attribution.\012",
			"<P>When you create an account, the form clears so you can create\012",
			"another immediately.<HR>\012",
			"<FORM METHOD=\"POST\" ACTION=\"/Administration/AccountCreationPage\">\012"
		);
	} else {
		&SendToClient("<HEAD><TITLE>Create a newsreading account</TITLE></HEAD>\012<BODY>\012",
			"<H2>Enter personal information for your new newsreading account</H2>\012",
			"You must have a user name and password on this server in order to use it to read news.\012",
			"Your user name and password can be almost anything you want.\012",
			"This server must also know your real name and your complete e-mail address,\012",
			"and optionally, the name of the organization to which you belong, so that articles\012",
			"you post have the proper attribution.<HR>\012",
			"Back to <A HREF=\"/WelcomePage\">welcome page</A><HR>\012",
			"<FORM METHOD=\"POST\" ACTION=\"/AccountCreationPage\">\012"
		);
	}
	
	local($InitialThe,$The);
	if ($Admin) {
		$InitialThe = "The";
		$The = "the";
	} else {
		$InitialThe = "Your";
		$The = "you";
	}
	
	if ($ExtraMessage ne "") {
		&SendToClient("<H3>$ExtraMessage</H3>\012");
	}
	
	&SendToClient(
		"Newsreading user name (at least $MinUserNameLength alphanumeric characters):<BR>\012",
			"<INPUT TYPE=\"text\" SIZE=20 NAME=\"username\" VALUE=\"$Username\"><BR>\012",
		"Newsreading password (at least $MinPasswordLength characters):<BR>\012",
			"<INPUT TYPE=\"password\" SIZE=20 NAME=\"password\" VALUE=\"$Password\"><BR>\012",
		"Retype $The password for verification:<BR>\012",
			"<INPUT TYPE=\"password\" SIZE=20 NAME=\"confirmpassword\" VALUE=\"$ConfirmPassword\"><BR>\012",
		"$InitialThe real name:<BR>\012",
			"<INPUT TYPE=\"text\" SIZE=40 NAME=\"realname\" VALUE=\"$Realname\"><BR>\012",
		"$InitialThe complete e-mail address:<BR>\012",
			"<INPUT TYPE=\"text\" SIZE=40 NAME=\"email\" VALUE=\"$Email\"><BR>\012",
		"$InitialThe organization (may be blank):<BR>\012",
			"<INPUT TYPE=\"text\" SIZE=40 NAME=\"organization\" VALUE=\"$Organization\"><BR>\012",
	);
	
	if ($Admin) {
		&SendToClient(
			"<P><INPUT TYPE=\"submit\" VALUE=\"Create account and clear form\"></FORM>\012",
			"Back to the <A HREF=\"/Administration/Index\">administration page</A><BR>\012",
			"Back to <A HREF=\"/WelcomePage\">welcome page</A><HR>\012",
			"</BODY>\012"
		);
	} else {
		&SendToClient("<P><INPUT TYPE=\"submit\" VALUE=\"Create account and start reading news\"></FORM>\012",
			"(you will be asked to type your user name and password again after you click this button)\012",
			"</BODY>\012"
		);
	}
}

# SendAdministrationPage($ProtoVers)
# Send a page that lets the administrator of the server do administrator-type
# things.

sub SendAdministrationPage {
	local($ProtoVers) = @_;

	&SendProtoVers($ProtoVers,$BuiltInExpirationSeconds);
	
	&SendToClient(
		"<HEAD><TITLE>Administration page</TITLE></HEAD>\012",
		"<BODY><H3>Administration page</H3>",
		"<UL>\012",
		"<LI><A HREF=\"/Administration/AccountCreationPage\">Create user accounts</A></LI>\012",
		"<LI><A HREF=\"/Administration/PasswordPage\">Change user account passwords</A></LI>\012",
		"<LI><A HREF=\"/Administration/Quit\">Shut down server</A></LI>\012",
		"</UL>\012",
		"<P>Back to <A HREF=\"/WelcomePage\">welcome page</A>\012",
		"</BODY>\012"
	);
}

# SendPasswordPage($ProtoVers,$Username,$ExtraMessage)
# Sends a page that lets the administrator change user passwords.
# $Username is a user name to fill into the form; it may be blank.
# $ExtraMessage is an extra message to print at the top of the form; it may be blank.

sub SendPasswordPage {
	local($ProtoVers,$Username,$ExtraMessage) = @_;
	
	&SendProtoVers($ProtoVers,$BuiltInExpirationSeconds);
	
	&SendToClient(
		"<HEAD><TITLE>Change account passwords</TITLE></HEAD>\012",
		"<BODY><H3>Change account passwords</H3>",
		$ExtraMessage,"\012",
		"<FORM METHOD=\"POST\" ACTION=\"/Administration/PasswordPage\">\012",
		"Account:<BR>\012",
			"<INPUT TYPE=\"text\" SIZE=20 NAME=\"username\" VALUE=\"$Username\"><BR>\012",
		"New password (at least $MinPasswordLength characters):<BR>\012",
			"<INPUT TYPE=\"password\" SIZE=20 NAME=\"password\" VALUE=\"\"><BR>\012",
		"Retype the password for verification:<BR>\012",
			"<INPUT TYPE=\"password\" SIZE=20 NAME=\"confirmpassword\" VALUE=\"\"><BR>\012",
		"<P><INPUT TYPE=\"submit\" VALUE=\"Change password and clear form\"></FORM>\012",
		"Back to the <A HREF=\"/Administration/Index\">administration page</A><BR>\012",
		"Back to <A HREF=\"/WelcomePage\">welcome page</A><HR>\012",
		"</BODY>\012"
	);
}

# SendHomeNewsPage($ProtoVers)
# This routine sends the news home page for the current user

sub SendHomeNewsPage {
	local($ProtoVers) = @_;
	
	&SendProtoVers($ProtoVers,$HomeExpirationSeconds);
	
	local($RealNamePossessive) = &PossessifyName($Newsrc{"prefs","realname"});
	&SendToClient("<HEAD><TITLE>",
			"$RealNamePossessive news home page",
			"</TITLE></HEAD>\012",
			"<BODY>\012",
			"<H2>$RealNamePossessive news home page</H2>\012"
	);
	
	# Clear and recalculate the cached subscription info.
	&ClearSubscriptionCache();
	&UpdateSubscriptionInfo();
	
	local($SubscriptionInfo) = "";
	
	if (@SortedSubscribedGroups > 0) {
		local($SubscribedGroup,$NumUnread);
		foreach $SubscribedGroup (@SortedSubscribedGroups) {
			$NumUnread = $SubscribedGroups{$SubscribedGroup};
			
			# Display it if it has articles, or if we're displaying every group anyway.
			if ($NumUnread > 0 || !$Newsrc{"prefs","hideempty"}) {
				local($ServerName,$GroupName) = split(/ /,$SubscribedGroup);
				$SubscriptionInfo .= "<B>" if ($NumUnread > 0);
				$SubscriptionInfo .= "<LI><A HREF=\"/Newsreading/NewsGroupPages/$ServerName/$GroupName\">$GroupName</A> $NumUnread unread</LI>\012";
				$SubscriptionInfo .= "</B>" if ($NumUnread > 0);
			}
		}
		
		if ($SubscriptionInfo ne "") {
			&SendToClient("Subscribed newsgroup",@SortedSubscribedGroups==1 ? "" : "s",":<BR><UL>\012",$SubscriptionInfo,"</UL>");
		} elsif (@SortedSubscribedGroups == 1) {
			&SendToClient("<B>Your one subscribed group is empty.</B><BR>\012");
		} else {
			&SendToClient("<B>Your ",@SortedSubscribedGroups+0," subscribed groups are all empty.</B>\012");
		}
	} else {
		&SendToClient("You are not subscribed to any groups. To subscribe to a group,\012",
				"either find it in the list of all newsgroups or\012",
				"type its name into the <A HREF=\"/Newsreading/SubscribePages\">mass-subscribe page</A>.\012"
		);
	}
	
	&SendToClient("<HR>\012");
	
	local($Server);
	local($ServerAlias);
	foreach $Server (@CanonicalNNTPServers) {
		$ServerAlias = &LongServerAlias($Server);
		&SendToClient("<A HREF=\"/Newsreading/AllGroupsPages/$Server/\">All newsgroups</A> on $ServerAlias<BR>\012");
	}
	
	&SendToClient(
			"<HR><A HREF=\"/Newsreading/PrefsPages\">Personal information page</A><BR>\012",
			"<A HREF=\"/Newsreading/KillPages\">Highlights and kills</A><BR>\012",
			"<A HREF=\"/Newsreading/SubscribePages\">Mass-subscribe to newsgroups</A><BR><BR>\012",
			"Back to <A HREF=\"/WelcomePage\">welcome page</A>\012",
			$GatewayVersion,
			"</BODY>\012"
	);
}

# SendSubscribePage($ProtoVers)

sub SendSubscribePage {
	local($ProtoVers) = @_;
	
	&SendProtoVers($ProtoVers,$BuiltInExpirationSeconds);
	
	&SendToClient(
			"<HEAD><TITLE>Mass-subscribe to groups for \"$CurrentUser\"</TITLE></HEAD>\012",
			"<BODY><H2>Mass-subscribe to newsgroups</H2>\012",
			"Back to <A HREF=\"/Newsreading/HomeNewsPage\">news home page</A><HR>\012",
			"There are two ways to subscribe to newsgroups. You can visit the groups individually and\012",
			"click their Subscribe buttons, or you can type or paste their names into this box\012",
			"and click the \"Subscribe and go to news home page\" button.<P>\012",
			"Lines containing ! are ignored. This means that if you have a .newsrc file\012",
			"in the usual unix format, you can copy your .newsrc and paste it here to subscribe to\012",
			"all the same groups. It won't pay attention to what articles you've already read, though.\012",
			"<FORM METHOD=\"POST\" ACTION=\"/Newsreading/SubscribePages\">\012",
			"<INPUT TYPE=\"submit\" VALUE=\"Subscribe and go to news home page\">\012",
			"<TEXTAREA cols=70 rows=10 NAME=\"groups\"></TEXTAREA></FORM></BODY>\012"
	);
}

# SendPrefsPage($ProtoVers,$ErrorMessage,$UseParamsInForm,$Password,$ConfirmPassword,$Realname,$Email,$Organization,
#				$TextSig,$HTMLSig,
#				$MarkRead,$HideEmpty,
#				$UseCheckboxes)

sub SendPrefsPage {
	local($ProtoVers,$ErrorMessage,$UseParamsInForm,$Password,$ConfirmPassword,$Realname,$Email,$Organization,
		  $TextSig,$HTMLSig,
		  $MarkRead,$HideEmpty,
		  $UseCheckboxes) = @_;
	
	&SendProtoVers($ProtoVers,$BuiltInExpirationSeconds);
	
	&SendToClient("<HEAD><TITLE>Personal information page for \"$CurrentUser\"</TITLE></HEAD>\012");
	if ($ErrorMessage ne "") {
		&SendToClient("<H1>$ErrorMessage</H1><HR>\012");
	}
	&SendToClient("<H2>Your personal information</H2><HR>\012");
	&SendToClient("Back to <A HREF=\"/Newsreading/HomeNewsPage\">news home page</A><HR>\012");
	
	local($PasswordFormValue,$ConfirmPasswordFormValue,$RealnameFormValue,$EmailFormValue,$OrganizationFormValue,
		  $TextSigFormValue,$HTMLSigFormValue,$MarkReadFormValue,$HideEmptyFormValue,$UseCheckboxesValue);
	if ($UseParamsInForm) {
		$PasswordFormValue = $Password;
		$ConfirmPasswordFormValue = $ConfirmPassword;
		$RealnameFormValue = $Realname;
		$EmailFormValue = $Email;
		$OrganizationFormValue = $Organization;
		$TextSigFormValue = $TextSig;
		$HTMLSigFormValue = $HTMLSig;
		$MarkReadFormValue = $MarkRead;
		$HideEmptyFormValue = $HideEmpty;
		$UseCheckboxesValue = $UseCheckboxes;
	} else {
		$PasswordFormValue = "";
		$ConfirmPasswordFormValue = "";
		$RealnameFormValue = $Newsrc{"prefs","realname"};
		$EmailFormValue = $Newsrc{"prefs","email"};
		$OrganizationFormValue = $Newsrc{"prefs","organization"};
		$TextSigFormValue = $Newsrc{"prefs","textsig"};
		$TextSigFormValue =~ s/\001/\012/g;
		$HTMLSigFormValue = $Newsrc{"prefs","htmlsig"};
		$HTMLSigFormValue =~ s/\001/\012/g;
		$MarkReadFormValue = $Newsrc{"prefs","markread"};
		$HideEmptyFormValue = $Newsrc{"prefs","hideempty"};
		$UseCheckboxesValue = $Newsrc{"prefs","nochecks"} ne "1";
	}
	
	local($MarkChecked) = $MarkReadFormValue ? " CHECKED" : "";
	local($HideChecked) = $HideEmptyFormValue ? " CHECKED" : "";
	local($CheckChecked) = $UseCheckboxesValue ? " CHECKED" : "";
	&SendToClient(
		"<FORM METHOD=\"POST\" ACTION=\"/Newsreading/PrefsPages\">\012",
		"<INPUT TYPE=\"checkbox\" NAME=\"markread\"$MarkChecked>Mark articles read as you read them<BR>\012",
		"<INPUT TYPE=\"checkbox\" NAME=\"hideempty\"$HideChecked>Show subscribed groups only if they have unread articles<BR>\012",
		"<INPUT TYPE=\"checkbox\" NAME=\"usecheck\"$CheckChecked>Include checkboxes on newsgroup pages<BR><BR>\012",
		"New password (at least $MinPasswordLength characters):<BR>",
			"<INPUT TYPE=\"password\" SIZE=20 NAME=\"password\" VALUE=\"$PasswordFormValue\"><BR>\012",
		"Retype new password:<BR>",
			"<INPUT TYPE=\"password\" SIZE=20 NAME=\"confirmpassword\" VALUE=\"$ConfirmPasswordFormValue\"><BR>\012",
		"Your real name:<BR>",
			"<INPUT TYPE=\"text\" SIZE=40 NAME=\"realname\" VALUE=\"$RealnameFormValue\"><BR>\012",
		"Your complete e-mail address:<BR>",
			"<INPUT TYPE=\"text\" SIZE=40 NAME=\"email\" VALUE=\"$EmailFormValue\"><BR>\012",
		"Name of your organization (may be blank):<BR>",
			"<INPUT TYPE=\"text\" SIZE=40 NAME=\"organization\" VALUE=\"$OrganizationFormValue\"><BR><BR>\012",
		"Signature for text postings:<BR>",
			"<TEXTAREA cols=70 rows=3 NAME=\"textsig\">$TextSigFormValue</TEXTAREA><BR>\012",
		"Signature for HTML postings:<BR>",
			"<TEXTAREA cols=70 rows=3 NAME=\"htmlsig\">$HTMLSigFormValue</TEXTAREA><BR>\012",
		"<INPUT TYPE=\"submit\" VALUE=\"Save settings and go to news home page\"></FORM>\012"
	);
}

# SendKillPage($ProtoVers)
# Send the user's highlight and kill information for perusal and zapping.
# NOTE: Entries in %KillFile which are neither highlights nor kills had
# better not match the regular expressions here.

sub SendKillPage {
	local($ProtoVers) = shift(@_);
	
	&SendProtoVers($ProtoVers,$KillExpirationSeconds);
	
	local($RealNamePossessive) = &PossessifyName($Newsrc{"prefs","realname"});
	&SendToClient("<HEAD><TITLE>");
	&SendToClient("$RealNamePossessive highlights and kills");
	&SendToClient("</TITLE></HEAD>\012");
	&SendToClient("<BODY>\012");
	&SendToClient("<H2>$RealNamePossessive highlights and kills</H2>\012");
	&SendToClient("<HR>Back to <A HREF=\"/Newsreading/HomeNewsPage\">news home page</A>\012");
	
	# Collect the highlights and kills into @Highlights and @Kills.
	local(@Highlights) = ();
	local(@Kills) = ();
	while (($Key,$Value) = each %KillFile) {
		if ($Key =~ /$;high....$;/) {
			push(@Highlights,$Key);
		} elsif ($Key =~ /$;kill....$;/) {
			push(@Kills,$Key);
		}
	}
	
	# Send out the highlights.
	if (@Highlights == ()) {
		&SendToClient("<HR><H3>You have no highlights.</H3>\012");
	} else {
		&SendToClient("<HR><H3>Highlights</H3>\012");
		&SendToClient("<FORM METHOD=\"POST\" ACTION=\"/Newsreading/KillPages\">\012");
		@Highlights = sort ByDate @Highlights;
		local($N) = 1;
		foreach $Item (@Highlights) {
			if ($Item !~ /^[^ ]+ ([^$;]+)$;high(....)$;(.*)$/) {
				print STDERR "Highlight '$Item' match unexpectedly failed in &SendKillPage.\n";
			} else {
				&SendToClient("<INPUT TYPE=\"checkbox\" NAME=\"high$N\" VALUE=\"$Item\">\012");
				$N += 1;
				local($Newsgroup) = $1;
				local($AuthSubj) = ($2 eq "Auth") ? "by" : "with subject";
				local($WhoWhat) = $3;
				local($Date) = &DateToString($KillFile{$Item});
				&SendToClient("articles $AuthSubj <B>$WhoWhat</B> in <B>$Newsgroup</B> (added $Date)<BR>\012");
			}
		}
		&SendToClient("<INPUT TYPE=\"hidden\" NAME=\"action\" VALUE=\"highlights\">\012");
		&SendToClient("<INPUT TYPE=\"submit\" VALUE=\"Remove Checked Highlights\"></FORM>\012");
	}
	
	# Send out the kills.
	if (@Kills == ()) {
		&SendToClient("<HR><H3>You have no kills.</H3>\012");
	} else {
		&SendToClient("<HR><H3>Kills</H3>\012");
		&SendToClient("<FORM METHOD=\"POST\" ACTION=\"/Newsreading/KillPages\">\012");
		@Kills = sort ByDate @Kills;
		local($N) = 1;
		foreach $Item (@Kills) {
			if ($Item !~ /^[^ ]+ ([^$;]+)$;kill(....)$;(.*)$/) {
				print STDERR "Kill '$Item' match unexpectedly failed in &SendKillPage.\n";
			} else {
				&SendToClient("<INPUT TYPE=\"checkbox\" NAME=\"kill$N\" VALUE=\"$Item\">\012");
				$N += 1;
				local($Newsgroup) = $1;
				local($AuthSubj) = ($2 eq "Auth") ? "by" : "with subject";
				local($WhoWhat) = $3;
				local($Date) = &DateToString($KillFile{$Item});
				&SendToClient("articles $AuthSubj <B>$WhoWhat</B> in <B>$Newsgroup</B> (added $Date)<BR>\012");
			}
		}
		&SendToClient("<INPUT TYPE=\"hidden\" NAME=\"action\" VALUE=\"kills\">\012");
		&SendToClient("<INPUT TYPE=\"submit\" VALUE=\"Remove Checked Kills\"></FORM>\012");
	}
	
	&SendToClient("<HR>Back to <A HREF=\"/Newsreading/HomeNewsPage\">news home page</A>\012");
	&SendToClient("</BODY>\012");
}

# ByDate()
# This is a sort subroutine; its parameters are passed in as $a and $b (see _Programming Perl_ p183).
# It orders two highlight or two kill items by the date they have in the %KillFile array.

sub ByDate {
	$KillFile{$a} cmp $KillFile{$b};
}

# DateToString(&Date)
# Convert a binary date (in the form returned by time()) to a string,
# for human reading.

sub DateToString {
	local($Date) = @_;
	local($Sec,$Min,$Hour,$MDay,$Mon,$Year,$WDay,$YDay,$IsDst) = localtime($Date);
	local(@MonthName) = ("Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec");
	$MDay." ".($MonthName[$Mon])." ".$Year;
}

# SendAllGroupsPage($ProtoVers,$GroupPrefix)
# This routine sends back an HTML page containing URLs for all the subfolders
# of $GroupPrefix on the current server.

sub SendAllGroupsPage {
	local($ProtoVers,$GroupPrefix) = @_;
	local($ServerAlias) = &ServerAlias($CurrentNNTPServer);
	
	&SendProtoVers($ProtoVers,$AllGroupsExpirationSeconds);
	&SendToClient("<HEAD><TITLE>");
	if ($GroupPrefix ne "") {
		&SendToClient("\"$GroupPrefix\" newsgroups on $ServerAlias");
	} else {
		&SendToClient("All newsgroups on $ServerAlias");
	}
	&SendToClient("</TITLE></HEAD>\012");
	&SendToClient("<BODY>\012");
	if ($GroupPrefix ne "") {
		&SendToClient("<H2>\"$GroupPrefix\" newsgroups on $ServerAlias</H2><HR>");
	} else {
		&SendToClient("<H2>All newsgroups on $ServerAlias</H2><HR>");
	}
	
	local($HeaderLinks) = "";
	$HeaderLinks .= "Back to <A HREF=\"/Newsreading/HomeNewsPage\">news home page</A>\012";
	if ($GroupPrefix ne "") {
		local($UpOneLevelGroupPrefix) = &UpOneLevelInAllGroups($GroupPrefix,0);
#		print STDERR "Up one level is [$UpOneLevelGroupPrefix]\n";
		if ($UpOneLevelGroupPrefix ne "") {
			$HeaderLinks .= "<BR>Back to <A HREF=\"/Newsreading/AllGroupsPages/$CurrentNNTPServer/$UpOneLevelGroupPrefix\">\"$UpOneLevelGroupPrefix\" newsgroups</A>\012";
		} else {
			$HeaderLinks .= "<BR>Back to <A HREF=\"/Newsreading/AllGroupsPages/$CurrentNNTPServer/\">all newsgroups</A>\012";
		}
	}
	&SendToClient($HeaderLinks);
	&SendToClient("<HR><UL>\012");
	local(@Subfolders) = split('/',$AllGroups{$GroupPrefix},100000); # the 100000 prevents trailing null fields from being stripped
	
	if (@Subfolders > 0) {
		shift(@Subfolders); 	# get rid of the extra empty string at the beginning
		@Subfolders = sort NewsgroupNameCompare @Subfolders;
		local($Subfolder);
		local(@NextLevelSubfolders);
		local($NextLevelSubfolder);
		foreach $Subfolder (@Subfolders) {
			if ($Subfolder eq "") {
				&SendToClient("<LI><A HREF=\"/Newsreading/NewsGroupPages/$CurrentNNTPServer/$GroupPrefix\">$GroupPrefix</A></LI>\012");
			} else {
				for (;;) {
					local($MaybeDottedGroupPrefix) = $GroupPrefix;
					($GroupPrefix ne "") && ($MaybeDottedGroupPrefix .= '.');
					
					@NextLevelSubfolders = split('/',$AllGroups{$MaybeDottedGroupPrefix.$Subfolder},3);					
					shift(@NextLevelSubfolders); 	# get rid of the extra empty string at the beginning
					if (@NextLevelSubfolders == 0) {
						# This should never happen - if there was a slash, then split will give you at least
						# two components.
						print STDERR "The impossible happened in SendAllGroupsPage.\n";
						last;
					} elsif (@NextLevelSubfolders == 1) {
						if ($NextLevelSubfolders[0] eq "") {
							# Then $Subfolder is a group
							&SendToClient("<LI><A HREF=\"/Newsreading/NewsGroupPages/$CurrentNNTPServer/$MaybeDottedGroupPrefix$Subfolder\">$MaybeDottedGroupPrefix$Subfolder</A></LI>\012");
							last;
						} else {
							# Then $Subfolder is a folder with only one subfolder, so we want to show that subfolder
							$Subfolder .= '.'.$NextLevelSubfolders[0];
							next;
						}
					} else {
						# Then $Subfolder is a folder with more than one subfolder, so we don't want to show its contents.
						local($SlashedSubfolder) = $MaybeDottedGroupPrefix.$Subfolder;
						$SlashedSubfolder =~ s#\.#/#g;
						&SendToClient("<LI><A HREF=\"/Newsreading/AllGroupsPages/$CurrentNNTPServer/$SlashedSubfolder\"><B>$MaybeDottedGroupPrefix$Subfolder...</B></A></LI>\012");
						last;
					}
				}
			}
		}
	}
	&SendToClient("</UL><HR>\012");
	&SendToClient($HeaderLinks);
	&SendToClient("</BODY>\012");
}

# UpOneLevelInAllGroups($GroupPrefix,$IsGroupName)
# Returns a string of the group prefix one level up in the current all groups tree
# from the given prefix.  Goes up more than one level if it encounters levels that
# only have one subfolder.  $IsGroupName should be true if the group prefix is to
# be interpreted as a group name, and false if it should be interpreted as a prefix in
# the all groups tree.  The difference being that, one level up from the group "alt.sex"
# is the "alt.sex" group prefix, but one level up from the "alt.sex" group prefix is the
# "alt" group prefix.

sub UpOneLevelInAllGroups {
	local($GroupPrefix,$IsGroupName) = @_;
	
	local(@UpOneLevelGroupPrefix) = split(/\./,$GroupPrefix);
	local(@Subgroups);
	while (@UpOneLevelGroupPrefix > 0) {
		pop(@UpOneLevelGroupPrefix) if (!$IsGroupName);
		@Subgroups = split('/',$AllGroups{join('.',@UpOneLevelGroupPrefix)},3);
		last if (@Subgroups > 2);
		pop(@UpOneLevelGroupPrefix) if ($IsGroupName);
	}
	
	join('.',@UpOneLevelGroupPrefix);
}

# SendNewsGroupPage($ProtoVers)
# This routine sends a page with URLs for all the unread articles in the current group.

sub SendNewsGroupPage {
	local($ProtoVers,$NNTPServer,$Group) = @_;
	
	local($NewsrcEntry) = $Newsrc{$NNTPServer." ".$Group};
	local($Subscribed) = 0;
	if ($NewsrcEntry ne "" && ($NewsrcEntry =~ /^:\s*([\d,\-]*)/)) {
		$Subscribed = 1;
	}
	
	# $NextGroup and $GoToNextGroup are used both here and at the end of this routine.
	# We need to calculate them before setting the server and group because &CalcGoToNextGroup
	# may change the server and group.
	local($NextGroup) = "";
	local($GoToNextGroup) = "";					# used here and at end of routine
	if ($Subscribed) {
		($NextGroup,$GoToNextGroup) = &CalcGoToNextGroup($Group);
	}
	
	&SetCurrentNNTPServer($NNTPServer,1);
	&SetCurrentGroup($Group,1);
	
	local($ServerAlias) = &ServerAlias($NNTPServer);
	
	local($HeaderLinks) = "Back to <A HREF=\"/Newsreading/HomeNewsPage\">news home page</A><BR>\012";
	local($UpOneLevelGroupPrefix) = &UpOneLevelInAllGroups($Group,1);
	if ($UpOneLevelGroupPrefix ne "") {
		$HeaderLinks .= "Back to <A HREF=\"/Newsreading/AllGroupsPages/$NNTPServer/$UpOneLevelGroupPrefix\">\"$UpOneLevelGroupPrefix\" newsgroups</A><BR>\012";
	} else {
		$HeaderLinks .= "Back to <A HREF=\"/Newsreading/AllGroupsPages/$NNTPServer/\">all newsgroups</A><BR>\012";
	}
	if ($GoToNextGroup ne "") {
		$HeaderLinks .= "Onward to $GoToNextGroup\012";
	}
	
	&SendProtoVers($ProtoVers,$NewsgroupExpirationSeconds);
	&SendToClient(
		"<HEAD><TITLE>\"$Group\" newsgroup on $ServerAlias</TITLE></HEAD>\012",
		"<BODY><H3>\"$Group\" newsgroup on $ServerAlias</H3><HR>\012",
		$HeaderLinks,
		"<HR>\012"
	);
	
	&MarkKilledArticlesRead();
	
	local($NumUnread) = &NumUnreadArticlesInGroup();
	
	&SendToClient(
			"<A HREF=\"/Newsreading/NewsGroupPages/$NNTPServer/$Group/MarkRead\">",
			"<IMG SRC=\"/Files/Buttons/mark-read.gif\" ALT=\"[Mark All Read]\"></A> ",
			
			"<A HREF=\"/Newsreading/NewsGroupPages/$NNTPServer/$Group/MarkUnread\">",
			"<IMG SRC=\"/Files/Buttons/mark-unread.gif\" ALT=\"[Mark All Unread]\"></A> ",
			
			$PostingAllowed ?
				"<A HREF=\"/Newsreading/PostArticlePages/$NNTPServer/$Group\">".
				"<IMG SRC=\"/Files/Buttons/post.gif\" ALT=\"[Post New Thread]\"></A> "
			:
				"",
			
			$Subscribed ?
				"<A HREF=\"/Newsreading/NewsGroupPages/$NNTPServer/$Group/Unsubscribe\">".
				"<IMG SRC=\"/Files/Buttons/unsubscribe.gif\" ALT=\"[Unsubscribe]\"></A>"
			:
				"<A HREF=\"/Newsreading/NewsGroupPages/$NNTPServer/$Group/Subscribe\">".
				"<IMG SRC=\"/Files/Buttons/subscribe.gif\" ALT=\"[Subscribe]\"></A>",
				
			"<BR>\012"
	);
	if (!$PostingAllowed) {
		&SendToClient("You are not allowed to post to this news server.<P>\012");
	}
	
	local($UseCheckboxes) = $Newsrc{"prefs","nochecks"} ne "1";
	
	if ($NumUnread > 0) {
		if ($NumUnread == 1) {
			&SendToClient("There is 1 unread article in this group.</A>\012");
		} else {
			&SendToClient("There are about $NumUnread unread articles in this group.</A>\012");
		}
		
		if ($UseCheckboxes) {
			&SendToClient("<FORM METHOD=\"POST\" ACTION=\"/Newsreading/NewsGroupPages/$NNTPServer/$Group\">\012");
		} else {
			&SendToClient("<UL>\012");
		}
		
		local(@Subjects) = split(/\t/,$ThreadDB{""});
		local($Subject);
		local($Root);
		local($PresentableSubject);
		local($Unread,$Highlighted);
		local($HStart,$HEnd);					# stuff to print for highlighted threads
		local($ThreadNum) = 0;
		
		foreach $Subject (@Subjects) {
			local(@Roots) = split(/\t/,$ThreadDB{"threadroots",$Subject});
			if (@Roots > 0) {
				$Root = $Roots[0];
				($Unread,$Highlighted) = &ScanThreadForInfo($Subject);
				if ($Unread > 1) {
					$HStart = $Highlighted ? "**<I>" : "";
					$HEnd = $Highlighted ? "</I>" : "";
					$PresentableSubject = &SubjectOfFullThread($Subject);
					
					if ($UseCheckboxes) {
						$ThreadNum += 1;
						&SendToClient(
							"<INPUT TYPE=\"checkbox\" NAME=\"T$ThreadNum\">\012",
							"<INPUT TYPE=\"hidden\" NAME=\"H$ThreadNum\" VALUE=\"$Subject\">\012",
							"$HStart<A HREF=\"/Newsreading/ThreadPages/$NNTPServer/$Group/$Subject\"><B>$PresentableSubject</B></A> $Unread articles",
							"$HEnd<BR>\012"
						);
					} else {
						&SendToClient(
							"<LI>$HStart<A HREF=\"/Newsreading/ThreadPages/$NNTPServer/$Group/$Subject\"><B>$PresentableSubject</B></A> $Unread articles",
							"$HEnd</LI>\012"
						);
					}
				} elsif ($Unread == 1) {
					&PrintThreadLine($Subject,$Highlighted);
				} else {
#					print STDERR "Thread has no unread articles.\n";
				}
			}
		}
		
		&SendToClient("</UL>\012") if (!$UseCheckboxes);
		
	} else {
		&SendToClient("<HR><H3>There are no unread articles in this group</H3>\012");
	}
	
	if ($NumUnread > 0) {
		&SendToClient("<HR>");
		if ($UseCheckboxes) {
			&SendToClient(
				"<INPUT TYPE=\"hidden\" NAME=\"Action\" Value=\"SkipUnchecked\">\012",
				"<INPUT TYPE=\"submit\" VALUE=\"Ignore Unchecked Articles\"></FORM>\012"
			);
		}
		&SendToClient("<FORM METHOD=\"POST\" ACTION=\"/Newsreading/NewsGroupPages/$NNTPServer/$Group\">\012");
		local($DestinationString,$Destination);
		if ($GoToNextGroup eq "") {
			$DestinationString = "Home";
			$Destination = "home";
		} else {
			local($ToServer,$ToGroup) = split(/ /,$NextGroup);
			$DestinationString = "To ".$ToGroup;
			$Destination = $NextGroup;
		}
		&SendToClient(
			"<INPUT TYPE=\"hidden\" NAME=\"Action\" Value=\"MarkAndGo\">\012",
			"<INPUT TYPE=\"hidden\" NAME=\"Destination\" Value=\"$Destination\">\012",
			"<INPUT TYPE=\"submit\" VALUE=\"Mark Articles Read and Go $DestinationString\"></FORM>\012",
			$HeaderLinks
		);
	}
	&SendToClient("<HR></BODY>\012");
}

# CalcGoToNextGroup($AfterThisGroup)
# Return a list, ($Group,HTML link), where $Group is the next subscribed newsgroup
# after $AfterThisGroup in the form "nntp.server news.group.name" and HTML link is
# an HTML link to that newsgroup. Return ("","") if there is no next subscribed group.

sub CalcGoToNextGroup {
	local($AfterThisGroup) = @_;
	local($State) = 0;
	local($Group);
	
	&UpdateSubscriptionInfo();
	
	foreach $Group (@SortedSubscribedGroups) {
		local($Server,$Newsgroup) = split(/ /,$Group);
		if ($State == 0) {
			$State = 1 if ($Newsgroup eq $AfterThisGroup);
		} elsif ($SubscribedGroups{$Group} > 0) {
			return ($Group,
				"<A HREF=\"/Newsreading/NewsGroupPages/$Server/$Newsgroup\">$Newsgroup</A> " .
				$SubscribedGroups{$Group} . " unread</A>");
		}
	}
	
	return ("","");
}

# PrintThreadLine($Subject,$Highlighted)
# Print one line about the given given thread, for the news groups page.
# The thread is known to have exactly one unread article.
# This is called only from &SendNewsGroupPage, which relies on this
# to "inherit" the local variables $UseCheckboxes, $ThreadNum and
# $Subject from &SendNewsGroupPage by dynamic scoping.

sub PrintThreadLine {
	local($Subject,$Highlighted) = @_;
	local(@Roots) = split(/\t/,$ThreadDB{"threadroots",$Subject});
	local($Root);
	
	foreach $Root (@Roots) {
		last if (&DoPrintThreadLine($Root,$Highlighted));
	}
}

# DoPrintThreadLine($Root)
# Recursive routine to do the work of PrintThreadLine.
# An invocation returns 1 if it found the article to print, so that
# its callers can stop looking for further articles.

sub DoPrintThreadLine {
	local($Root,$Highlighted) = @_;
	local($ArticleNum) = $ThreadDB{$Root,"artnum"};
	
	if ($ArticleNum ne "dummy" && !&HasArticleBeenRead($ArticleNum)) {
		local($Author,$RealSubject,$NumLines);
		$RealSubject = &SubjectOfArticle($Root);
		$Author = &AuthorOfArticle($Root);
		$NumLines = $ThreadDB{$Root,"numlines"};
		$NumLines = "???" if ($NumLines eq "");
		if ($UseCheckboxes) {
			$ThreadNum += 1;
			&SendToClient(
				"<INPUT TYPE=\"checkbox\" NAME=\"T$ThreadNum\">\012",
				"<INPUT TYPE=\"hidden\" NAME=\"H$ThreadNum\" VALUE=\"$Subject\">\012"
			);
		} else {
			&SendToClient("<LI>");
		}
		&SendToClient("**<I>") if ($Highlighted);
		&SendToClient("<A HREF=\"/Newsreading/ArticlePages/$NNTPServer/$Group/$Root\">$RealSubject, by $Author</A> $NumLines lines");
		&SendToClient("</I>") if ($Highlighted);
		if ($UseCheckboxes) {
			&SendToClient("<BR>\012");
		} else {
			&SendToClient("</LI>\012");
		}
		return 1;									# the article was found; stop and say so
	}
	
	local($Children) = $ThreadDB{$Root,"children"};
	local($Child);
	foreach $Child (split(/ /,$Children)) {
		&DoPrintThreadLine($Child,$Highlighted) && return 1;	# if we've found the article, stop looking
	}
	
	0;												# the article was not found here
}

# SendThreadPage($ProtoVers,$Subject)
# Sends a page with a list of all the articles in the thread in the current group.

sub SendThreadPage {
	local($ProtoVers,$Subject) = @_;
	
	&SendProtoVers($ProtoVers,$NewsgroupExpirationSeconds);
	
	$PresentableSubject = &SubjectOfFullThread($Subject);
	
	&SendToClient("<HEAD><TITLE>");
	&SendToClient("\"$PresentableSubject\" thread");
	&SendToClient("</TITLE></HEAD>\012");
	&SendToClient("<BODY>\012");

	&SendToClient("<H2>\"$PresentableSubject\" thread</H2><HR>\012");
	local($HeaderLinks) = "";
	$HeaderLinks .= "Back to <A HREF=\"/Newsreading/HomeNewsPage\">news home page</A><BR>\012";
	$HeaderLinks .= "Back to <A HREF=\"/Newsreading/NewsGroupPages/$CurrentNNTPServer/$CurrentGroup\">\"$CurrentGroup\" newsgroup</A>\012";
	&SendToClient("$HeaderLinks<HR><UL>\012");
	&PrintOneThreadHTML($Subject);
	&SendToClient("</UL>\012");
	&SendToClient("<HR>$HeaderLinks</BODY>\012");
}

# PrintOneThreadHTML($Subject)
# Finds all the roots of the given thread and calls DoPrintOneThreadHTML
# to print out the thread tree recursively.

sub PrintOneThreadHTML {
	local($Subject) = @_;
	local(@Roots) = split(/\t/,$ThreadDB{"threadroots",$Subject});
	local($Root);
	
	foreach $Root (@Roots) {
		&DoPrintOneThreadHTML($Root);
	}
}

# DoPrintOneThreadHTML($RootMsgID)
# This routine prints out HTML code for the thread with the given root.

sub DoPrintOneThreadHTML {
	local($Root) = @_;
	local($ArticleNum) = $ThreadDB{$Root,"artnum"};
	local($PrintedArticle) = 0;

	if ($ArticleNum ne "dummy" && !&HasArticleBeenRead($ArticleNum)) {
		local($Subject,$Author,$NumLines);
		local($Highlighted);
		$Subject = &SubjectOfArticle($Root);
		$Author = &AuthorOfArticle($Root);
		$NumLines = $ThreadDB{$Root,"numlines"};
		$NumLines = "???" if ($NumLines eq "");
		$Highlighted = &ArticleHighlighted($Root);
		&SendToClient("<LI>");
		&SendToClient("<I>") if ($Highlighted);
		&SendToClient("<A HREF=\"/Newsreading/ArticlePages/$CurrentNNTPServer/$CurrentGroup/$Root\">$Subject, by $Author</A> $NumLines lines");
		&SendToClient("</I>") if ($Highlighted);
		&SendToClient("</LI>\012");
		$PrintedArticle = 1;
	} else {
#		&SendToClient("<LI>Dummy article $Root</LI>\012");
#		$PrintedArticle = 1;
	}
	
	local($Children) = $ThreadDB{$Root,"children"};
	local($Child);
	foreach $Child (split(/ /,$Children)) {
		if ($PrintedArticle) {
			&SendToClient("<UL>\012");
		}
		&DoPrintOneThreadHTML($Child);
		if ($PrintedArticle) {
			&SendToClient("</UL>\012");
		}
	}
}

# AuthorOfArticle($MsgID)
# Returns the author of the article, or "[no author]" if appropriate, in HTMLSafeified form.

sub AuthorOfArticle {
	local($MsgID) = @_;
	
	$Author = $ThreadDB{$MsgID,"authorreal"};
	$Author = $ThreadDB{$MsgID,"authoremail"} if ($Author eq "");
	if ($Author eq "") {
		$Author = "[no author]";
	} else {
		&HTMLSafeify($Author);
	}
}

# Subjectify($Subject)
# Convert a subject string into user-safe and HTML-safe form.
# An empty subject becomes "[no subject]".

sub Subjectify {
	local($Subject) = @_;
	
	if ($Subject eq "") {
		$Subject = "[no subject]";
	} else {
		$Subject = &HTMLSafeify($Subject);
	}
	
	$Subject;
}

# SubjectOfArticle($MsgID)
# Returns the subject of the article, or "[no subject]" if appropriate, in HTMLSafeified form.

sub SubjectOfArticle {
	local($MsgID) = @_;
	
	&Subjectify($ThreadDB{$MsgID,"subject"});
}

# SubjectOfFullThread($Subject)
# Given the canonical subject of a thread, returns the subject of the thread, or "[no subject]"
# if appropriate, in HTMLSafeified form.

sub SubjectOfFullThread {
	local($Subject) = @_;
	
	&Subjectify($ThreadDB{"threadsubject",$Subject});
}

# DoSubjectOfThread($RootMsgID)
# This routine returns the subject of the first non-dummy article it encounters in the given
# subthread's tree (going in depth first order).

sub DoSubjectOfThread {
	local($RootMsgID) = @_;
	local($ArticleNum) = $ThreadDB{$RootMsgID,"artnum"};
	local($Result) = "";
	
	if ($ArticleNum ne "dummy") {
		$Result = $ThreadDB{$RootMsgID,"subject"};
	} else {
		local($Children) = $ThreadDB{$RootMsgID,"children"};
		local($Child);
		foreach $Child (split(/ /,$Children)) {
			$Result = &DoSubjectOfThread($Child);
			last if ($Result ne "");
		}
	}
	
	$Result;
}

# NumUnreadArticlesInThread($Subject)
# Finds all the roots of the given thread and calls DoNumUnreadArticlesInThread
# to find the number of unread articles recursively.

sub NumUnreadArticlesInThread {
	local($Subject) = @_;
	local(@Roots) = split(/\t/,$ThreadDB{"threadroots",$Subject});
	local($Root);
	local($NumUnread) = 0;
	
	foreach $Root (@Roots) {
		$NumUnread += &DoNumUnreadArticlesInThread($Root);
	}
	
	$NumUnread;
}

# DoNumUnreadArticlesInThread($RootMsgID)
# This routine returns the number of unread articles in the given thread.

sub DoNumUnreadArticlesInThread {
	local($RootMsgID) = @_;
	local($ArticleNum) = $ThreadDB{$RootMsgID,"artnum"};
	local($Children) = $ThreadDB{$RootMsgID,"children"};
	local($NumUnread) = 0;

	if ($ArticleNum ne "dummy" && !&HasArticleBeenRead($ArticleNum)) {
		$NumUnread++;
	}
	
	local($Child);
	foreach $Child (split(/ /,$Children)) {
		$NumUnread += &DoNumUnreadArticlesInThread($Child);
	}
	
	$NumUnread;
}

# ScanThreadForInfo($Subject) -> ($NumUnread,$Highlighted)
# Returns a list (number of unread articles in thread,whether thread is highlighted).

sub ScanThreadForInfo {
	local($Subject) = @_;
	local(@Roots) = split(/\t/,$ThreadDB{"threadroots",$Subject});
	local($Root);
	local($NumUnread) = 0;
	local($Highlighted) = 0;
	local($Unread);
	
	foreach $Root (@Roots) {
		if ($Highlighted) {
			$NumUnread += &DoNumUnreadArticlesInThread($Root);			# no need to check for further highlights
#			print STDERR "Switched over; NumUnread is up to $NumUnread\n";
		} else {
			($Unread,$Highlighted) = &DoScanThreadForInfo($Root);
			$NumUnread += $Unread;
#			print STDERR "NumUnread is up to $NumUnread\n";
		}
	}
	
	($NumUnread,$Highlighted);
}

# DoScanThreadForInfo($RootMsgID)
# Returns a list (number of unread articles in thread,whether thread is highlighted),
# for a given subthread.

sub DoScanThreadForInfo {
	local($RootMsgID) = @_;
	local($ArticleNum) = $ThreadDB{$RootMsgID,"artnum"};
	local($Children) = $ThreadDB{$RootMsgID,"children"};
	local($NumUnread) = 0;
	local($Highlighted) = 0;
	local($Unread);
	
	if ($ArticleNum ne "dummy" && !&HasArticleBeenRead($ArticleNum)) {
		$NumUnread++;
		$Highlighted = &ArticleHighlighted($RootMsgID);
	}
	
	local($Child);
	foreach $Child (split(/ /,$Children)) {
		if ($Highlighted) {
			$NumUnread += &DoNumUnreadArticlesInThread($Child);			# no need to check for further highlights
		} else {
			($Unread,$Highlighted) = &DoScanThreadForInfo($Child);
			$NumUnread += $Unread;
		}
	}
	
	($NumUnread,$Highlighted);
}

# RootOfThread($MsgID)
# Returns the message id of the root of the thread in which the given message resides.
# $$$ is this correct for disjoint threads? no; causes bug in which Prev & Next links don't appear

sub RootOfThread {
	local($MsgID) = @_;
	
	while ($ThreadDB{$MsgID,"parent"} ne "") {
		$MsgID = $ThreadDB{$MsgID,"parent"};
	}
	
	$MsgID;
}

# GetArticleHeaderAndBody($MessageID) -> ($ArticleHeaders, $ArticleBody)
# Fetch the given article from the current NNTP server and split it
# into header and body portions.

sub GetArticleHeaderAndBody {
	local($MessageID) = @_;
	
	local(@ArticleLines) = split(/\015\012/,&GetArticleText($MessageID),100000);
	local($ArticleNum) = $ThreadDB{$MessageID,"artnum"};
	
	local($Headers) = "";
	local($Body) = "";
	local($Line);
	local($DoneWithHeader) = 0;
	foreach $Line (@ArticleLines) {
		if ($Line eq "" && !$DoneWithHeader) {
			$DoneWithHeader = 1;
			next;
		}
		if ($DoneWithHeader) {
			$Body .= $Line."\012";
		} else {
			$Headers .= $Line."\015\012";
		}
	}
	
	($Headers, $Body);
}

# AuthorEmailString($MessageID,$MailtoURL)
# Return a string to represent the author of an article, in as nice a form as possible.
# If $MailtoURL is true and the author's e-mail address is known, return a mailto: URL.
# The possible forms, in order of niceness, are, if $MailtoURL is true:
#	<A HREF="mailto:email@address">Author Name (email@address)</A>
#	<A HREF="mailto:email@address">email@address</A>
#	Author Name
#	[unknown author]
# If $MailtoURL is false, they are:
#	Author Name (email@address)
#	email@address
#	Author Name
#	[unknown author]

sub AuthorEmailString {
	local($MessageID,$MailtoURL) = @_;
	
	local($Author) = $ThreadDB{$MessageID,"authorreal"};
	local($Email) = $ThreadDB{$MessageID,"authoremail"};
	if ($Author eq "" && $Email eq "") {
		$Author = "[unknown author]";
	} elsif ($Author eq "") {
		$Author = $Email;
	} elsif ($Email eq "") {
		# no change
	} else {
		$Author .= " (".$Email.")";
	}
	&HTMLSafeify($Author);
	if ($MailtoURL && $Email ne "") {
		$Author = "<A HREF=\"mailto:$Email\">$Author</A>";
	}
	
	$Author;
}

# SendArticlePage($ProtoVers,$MessageID)
# Sends a page with the given article (and appropriate links) on it.

sub SendArticlePage {
	local($ProtoVers,$MessageID) = @_;
	
	local($ArticleHeader,$ArticleBody) = &GetArticleHeaderAndBody($MessageID);
	local(%ArticleHeaders) = &BuildHeaderArray($ArticleHeader);
	
	# Capture a copy of the article for debugging.
#	print STDERR "$ArticleHeader\n\n$ArticleBody\n";
#	open(OUTFILE,">out");
#	print OUTFILE $ArticleBody;
#	close OUTFILE;
	
	local($ReallyPlainText) = 0;
	local($ReallyBase64) = 0;
	
	if (!$ArticleHeaders{"date"}) {
		$ArticleHeaders{"date"} = &TimeString(time);
	}
	if (!$ArticleHeaders{"mime-version"}) {
		$ArticleHeaders{"mime-version"} = "1.0";
	}
	if ((!$ArticleHeaders{"content-type"}) || ($ArticleHeaders{"content-type"} =~ /^text($|\/plain)/i)) {
		$ArticleHeaders{"content-type"} = "text/html";
		$ReallyPlainText = 1;
	}
	if ($ArticleHeaders{"content-transfer-encoding"} =~ /base64/i) {
		$ArticleHeaders{"content-transfer-encoding"} = "8bit";
		$ReallyBase64 = 1;
	}
	
	if ($ProtoVers ne "") {
		&SendToClient(
			"HTTP/1.0 200 OK\012",
			"Date: ",$ArticleHeaders{"date"},"\012",
			"Expires: ",&TimeString(time+$ArticleExpirationSeconds),"\012",
			"Mime-Version: ",$ArticleHeaders{"mime-version"},"\012",
			"Content-Type: ",$ArticleHeaders{"content-type"},"\012",
			"Content-Transfer-Encoding: ",$ArticleHeaders{"content-transfer-encoding"},"\012\012"
		);
	}
	local($Root);
	local($HeaderLinks) = "";
	if ($ReallyPlainText || $ArticleHeaders{"content-type"} =~ /text\/html/i) {
		&SendToClient("<HEAD><TITLE>");
		&SendToClient("Article &lt;$MessageID&gt; in \"$CurrentGroup\" newsgroup");
		&SendToClient("</TITLE></HEAD><BODY>\012");
		$Root = &RootOfThread($MessageID);
		$HeaderLinks .= "See <A HREF=\"/Newsreading/TextArticlePages/$CurrentNNTPServer/$CurrentGroup/$MessageID\">this article in text form</A>, with full headers<BR>\012";
		$HeaderLinks .= "Back to <A HREF=\"/Newsreading/HomeNewsPage\">news home page</A><BR>\012";
		if ($CurrentGroup ne "") {
			$HeaderLinks .= "Back to <A HREF=\"/Newsreading/NewsGroupPages/$CurrentNNTPServer/$CurrentGroup\">\"$CurrentGroup\" newsgroup</A><BR>\012";
			local($ThreadSubject) = &DoSubjectOfThread($Root);
			local($CanonicalSubject) = &CanonicalizeSubject($ThreadSubject);
			local($Prev,$Next);
			if (&NumUnreadArticlesInThread($CanonicalSubject) > 1) {
				$HeaderLinks .= "Back to <A HREF=\"/Newsreading/ThreadPages/$CurrentNNTPServer/$CurrentGroup/$CanonicalSubject\">\"$ThreadSubject\" thread</A><BR>\012";
				($Prev,$Next) = &PreviousAndNextArticles($CanonicalSubject,$MessageID);
				if ($Prev ne "") {
					$HeaderLinks .= "Previous <A HREF=\"/Newsreading/ArticlePages/$CurrentNNTPServer/$CurrentGroup/$Prev\">".&AuthorOfArticle($Prev)."</A>";
				}
				$HeaderLinks .= " -- ";
				if ($Next ne "") {
					$HeaderLinks .= "Next <A HREF=\"/Newsreading/ArticlePages/$CurrentNNTPServer/$CurrentGroup/$Next\">".&AuthorOfArticle($Next)."</A>";
				}
				$HeaderLinks .= "<BR>\012";
			}
		}
		
		if ($PostingAllowed) {
			if (&ValidEmailAddress($Newsrc{"prefs","email"}) && $Newsrc{"prefs","realname"} ne "") {
				$HeaderLinks .= "<FORM METHOD=\"GET\" ACTION=\"/Newsreading/PostArticlePages/$CurrentNNTPServer/$CurrentGroup/$MessageID\">\012";
				$HeaderLinks .= "<INPUT TYPE=\"submit\" VALUE=\"Post Followup Article\">\012";
				$HeaderLinks .= "<INPUT TYPE=\"checkbox\" NAME=\"original\" CHECKED>Include original in followup</FORM>\012";
			} else {
				$HeaderLinks .= "You cannot post until you enter your name and e-mail address on your personal information page.\012";
			}
		} else {
			$HeaderLinks .= "You do not have permission to post articles to this news server.\012";
		}
		
		&SendToClient($HeaderLinks);
		
		&SendToClient("<HR><PRE>\012");
		
		local($Author) = &AuthorEmailString($MessageID,1);
		local($Subject) = &SubjectOfArticle($MessageID);
		
		&SendToClient("<B>Subject</B>:       $Subject\012");
		&SendToClient("<B>From</B>:          $Author\012");

		if ($ArticleHeaders{"organization"}) {
			&SendToClient("<B>Organization</B>: ",&HTMLSafeify($ArticleHeaders{"organization"}),"\012");
		}
		&SendToClient("<B>Date</B>:          ",&HTMLSafeify($ArticleHeaders{"date"}),"\012");
		
		# Calculate references and responses lines.
		local($RefsToSend) = "";
		local($KidsToSend) = "";
		if ($ArticleHeaders{"references"}) {
			local(@References) = reverse(&ValidReferences($ArticleHeaders{"references"}));
			local($Reference);
			foreach $Reference (@References) {
				if (&DoesArticleExist($Reference)) {
					local($RefAuthor) = &AuthorOfArticle($Reference);
					$RefsToSend .= "<A HREF=\"/Newsreading/ArticlePages/$CurrentNNTPServer/$CurrentGroup/$Reference\">$RefAuthor</A>, ";
				}
			}
		}
		local($Children) = $ThreadDB{$MessageID,"children"};
		if ($CurrentGroup ne "" && $Children ne "") {
			local(@Children) = split (/ /,$Children);
			local($Child);
			local($OneExisted) = 0;
			foreach $Child (@Children) {
				if (&DoesArticleExist($Child)) {
					local($RefAuthor) = &AuthorOfArticle($Child);
					$KidsToSend .= "<A HREF=\"/Newsreading/ArticlePages/$CurrentNNTPServer/$CurrentGroup/$Child\">$RefAuthor</A>, ";
				}
			}
		}
		
		# Send references and responses lines.
		if ($RefsToSend ne "" || $KidsToSend ne "") {
			&SendToClient("<HR>");
			if ($RefsToSend ne "") {
				$RefsToSend =~ s/\,\ $//;			# remove final ", "
				&SendToClient("References:   $RefsToSend\012");
			}
			if ($KidsToSend ne "") {
				$KidsToSend =~ s/\,\ $//;			# remove final ", "
				&SendToClient("Responses:    $KidsToSend\012");
			}
			&SendToClient("<HR>\012");
		}
		
		&SendToClient("</PRE>\012");
	}
	if ($ReallyPlainText) {
		&SendToClient("<PRE>\012");
		local($OldStringStar) = $*;
		$* = 1;
		# Convert strings that look like HTML links into real HTML links.
		# We change the < and > characters in the links into \001 and \002
		# temporarily, so that &HTMLSafeify does not replace them.
		# This method fails when both patterns match, which happens for strings
		# like this:
		#	<a href="http://www.telegraph.co.uk">http://www.telegraph.co.uk</a>
		$ArticleBody =~ s/\001//g;		# first remove any real \001 or \002 characters
		$ArticleBody =~ s/\002//g;
		$ArticleBody =~ s#<(a\s+href\s*=)(\s*("))?\s*([^>]+)>([^<]+)</a>#\001$1$3$4\002$5\001/A\002\012#gi;
		$* = $OldStringStar;
		$ArticleBody =~ s#(^|[^\w="])(http\:|ftp\:|gopher\:|telnet\:|news\:|finger\:)//([^\s<>"')]+)#$1\001A HREF="$2//$3"\002$2//$3\001/A\002#gi;
		$ArticleBody = &HTMLSafeify($ArticleBody);
		$ArticleBody =~ s/\001/</g;
		$ArticleBody =~ s/\002/>/g;
	}
	if ($ReallyBase64) {
		local($ArticleTextLine);
		@ArticleTextLines = split(/\012/,$ArticleBody,100000);
		foreach $ArticleTextLine (@ArticleTextLines) {
			&SendToClient(&Base64Decode($ArticleTextLine));
		}
	} else {
		&SendToClient($ArticleBody);
	}
	if ($ReallyPlainText) {
		&SendToClient("</PRE>\012");
	}
	if ($ReallyPlainText || $ArticleHeaders{"content-type"} =~ /^text\/html/i) {
		&SendToClient("<HR>\012");
		&SendToClient($HeaderLinks);
		if ($CurrentGroup ne "") {
			if (&ArticleHighlighted($MessageID) == 1) {
				# Highlighted by author, so make the default be Highlight Subject.
				$AuthorCheck = "";
				$SubjectCheck = " CHECKED";
			} else {
				# Not highlighted, or highlighted by subject. The default is Highlight Author.
				$AuthorCheck = " CHECKED";
				$SubjectCheck = "";
			}
			&SendToClient(
				"<HR>\012<FORM METHOD=\"POST\" ACTION=\"/Newsreading/ArticlePages/$CurrentNNTPServer/$CurrentGroup/$MessageID\">\012",
				"<INPUT TYPE=\"radio\" NAME=\"Highlight\" VALUE=\"Author\"$AuthorCheck>Author\012",
				"<INPUT TYPE=\"radio\" NAME=\"Highlight\" VALUE=\"Subject\"$SubjectCheck>Subject\012",
				"<INPUT TYPE=\"submit\" VALUE=\"Highlight\"></FORM><BR>\012",
			
				"<FORM METHOD=\"POST\" ACTION=\"/Newsreading/ArticlePages/$CurrentNNTPServer/$CurrentGroup/$MessageID\">\012",
				"<INPUT TYPE=\"radio\" NAME=\"Kill\" VALUE=\"Author\" CHECKED>Author\012",
				"<INPUT TYPE=\"radio\" NAME=\"Kill\" VALUE=\"Subject\">Subject\012",
				"<INPUT TYPE=\"submit\" VALUE=\"Kill\"></FORM><BR>\012"
			);
		}
		
		&SendToClient("</BODY>\012");
	}
	
	# If the markread option is turned on, mark the article read.
	# Do it last, in case there's an error before here.
	if ($CurrentGroup ne "" && $Newsrc{"prefs","markread"}) {
		local($ArticleNum) = $ThreadDB{$MessageID,"artnum"};
		&MarkArticleRead($ArticleNum);
	}
}

# PreviousAndNextArticles($MessageID)
# Return the message IDs of the previous and next articles in this thread.
# If there is no corresponding article, return "" instead of a message ID.

sub PreviousAndNextArticles {
	local($Subject,$This) = @_;
	local(@Roots) = split(/\t/,$ThreadDB{"threadroots",$Subject});
	local($Root);
	local($Old);
	local($Prev,$Next)=("","");
	
	# We only need to search three roots, previous, current, and next.
	# For now, we start at the beginning for simplicity.
	$Old = "";
	foreach $Root (@Roots) {
		($Old,$Prev,$Next) = &DoPreviousAndNextArticles($Root,$Old,$Prev,$This,$Next);
		last if ($Next ne "");
	}
	
	($Prev,$Next);
}

# DoPreviousAndNextArticles($Root,$Old,$Prev,$This,$Next)
# Recursively do the work of &PreviousAndNextArticles.
# $Root - the article now being examined
# $Old  - the previous article examined
# $Prev - the previous article, value so far (to be returned)
# $This - the article for which we're finding $Next and $Prev
# $Next - the next article, value so far (to be returned)

sub DoPreviousAndNextArticles {
	local($Root,$Old,$Prev,$This,$Next) = @_;
	
	local($ArticleNum) = $ThreadDB{$Root,"artnum"};
	
	if ($ArticleNum ne "dummy" && !&HasArticleBeenRead($ArticleNum)) {
		if ($Root eq $This) {
			$Prev = $Old;
		} elsif ($Old eq $This) {
			$Next = $Root;
			return ($Root,$Prev,$Next);
		}
		$Old = $Root;
	}
	
	local($Children) = $ThreadDB{$Root,"children"};
	local($Child);
	foreach $Child (split(/ /,$Children)) {
		($Old,$Prev,$Next) = &DoPreviousAndNextArticles($Child,$Old,$Prev,$This,$Next);
		last if ($Next ne "");
	}
	
	($Old,$Prev,$Next);
}

# SendTextArticlePage($ProtoVers,$MessageID)
# Sends a page with the given article in text form, with full headers.

sub SendTextArticlePage {
	local($ProtoVers,$MessageID) = @_;
	
	if ($ProtoVers ne "") {
		local($Now) = time;
		&SendToClient(
			"HTTP/1.0 200 OK\012",
			"Date: ",&TimeString($Now),"\012",
			"Expires: ",&TimeString($Now+$ArticleExpirationSeconds),"\012",
			"Mime-version: 1.0\012",
			"Content-type: text/plain\012",
			"\012"
		);
	}
	
	local($ArticleText) = &GetArticleText($MessageID);
	$ArticleText =~ s/\015\012/\012/g;
	&SendToClient($ArticleText);
	
	# If the markread option is turned on, mark the article read.
	# Do it last, in case there's an error before here.
	if ($CurrentGroup ne "" && $Newsrc{"prefs","markread"}) {
		&MarkArticleRead($ThreadDB{$MessageID,"artnum"});
	}
}

# SendPostArticlePage($ProtoVers,$MsgIDRespondingTo)
# Send a form that you can fill out to post an article.
# If $MsgIDRespondingTo is "", then the article starts a new thread.
# If $IncludeOriginal is true, then include the original article in the initial
# followup text, in the usual netnews way.

sub SendPostArticlePage {
	local($ProtoVers,$MsgIDRespondingTo,$IncludeOriginal) = @_;

	&SendProtoVers($ProtoVers,$BuiltInExpirationSeconds);
	
	if (!$PostingAllowed) {
		local($ServerAlias) = &LongServerAlias($CurrentNNTPServer);
		&SendToClient("<HEAD><TITLE>You cannot post to \"$ServerAlias\"</TITLE></HEAD>\012");
		&SendToClient("<BODY><H2>You do not have permission to post articles to $ServerAlias</H2></BODY>\012");
		return;
	}
	
	if (!&ValidEmailAddress($Newsrc{"prefs","email"}) || $Newsrc{"prefs","realname"} eq "") {
		&SendToClient("<HEAD><TITLE>You must enter your name and e-mail address</TITLE></HEAD>\012");
		&SendToClient("<BODY><H2>You are not allowed to post until you enter your real name and e-mail address on your personal information page.</H2></BODY>\012");
		return;
	}
	
	local($Author) = &AuthorEmailString($MsgIDRespondingTo,1);
	&SendToClient("<HEAD><TITLE>");
	if ($MsgIDRespondingTo ne "") {
		&SendToClient("Post response to article by $Author");
	} else {
		&SendToClient("Post article in new thread");
	}
	&SendToClient("</TITLE></HEAD>\012<BODY>\012");
	
	if ($MsgIDRespondingTo eq "") {
		&SendToClient("<H2>Post article in new thread</H2>\012");
	} else {
		&SendToClient("<H2>Post response to article by $Author</H2>\012");
	}
	
	# If we're including the original in the followup, figure out what the original is.
	local($ReplyBody) = "";
	if ($IncludeOriginal && $MsgIDRespondingTo ne "") {
		local($ArticleHeader,$ArticleBody) = &GetArticleHeaderAndBody($MsgIDRespondingTo);
		local(%ArticleHeaders) = &BuildHeaderArray($ArticleHeader);
		
		$ReplyBody = "In article <$MsgIDRespondingTo>, ".&AuthorEmailString($MsgIDRespondingTo,0)." wrote:\012";
		if ($ArticleHeaders{"content-transfer-encoding"} =~ /base64/i) {
			# There doesn't seem to be anything great to do about this case.
			$ReplyBody .= "[the article was not text]\012";
		} elsif ($ArticleBody ne "") {
			# If there's at least one line in the article body.
			$ReplyBody .= $ArticleBody;
			$ReplyBody =~ s/\012/\012>/g;					# may add an extra > to the end
			chop($ReplyBody) if ($ReplyBody =~ /\012>$/);	# strip off the extra >
		}
	}
	
	&SendToClient(
		"<HR>Back to <A HREF=\"/Newsreading/NewsGroupPages/$CurrentNNTPServer/$CurrentGroup\">\"$CurrentGroup\" newsgroup</A><HR>\012",
		"<FORM METHOD=\"POST\" ACTION=\"/Newsreading/PostArticlePages/$CurrentNNTPServer/$CurrentGroup/$MsgIDRespondingTo\">\012",
		"Newsgroups:<BR><INPUT TYPE=\"text\" SIZE=60 NAME=\"Newsgroups\" VALUE=\"$CurrentGroup\"><BR>\012",
#		"Followups to:<BR><INPUT TYPE=\"text\" SIZE=60 NAME=\"Followups\" VALUE=\"$CurrentGroup\"><BR>\012"
	);
	local($DefaultSubject) = "";
	if ($MsgIDRespondingTo ne "") {
		$DefaultSubject = &SubjectOfArticle($MsgIDRespondingTo);
		if ($DefaultSubject !~ /^re: /i) {
			$DefaultSubject = "Re: ".$DefaultSubject;
		}
	}
	&SendToClient("Subject:<BR><INPUT TYPE=\"text\" SIZE=60 NAME=\"Subject\" VALUE=\"$DefaultSubject\"><BR>\012");
#	&SendToClient("Distribution:<BR><SELECT NAME=\"Distribution\">\012");
#	&SendToClient("<OPTION>world<OPTION>local\012");
#	&SendToClient("</SELECT><BR>\012");
	&SendToClient("Article type:<BR><SELECT NAME=\"Contenttype\">\012");
	&SendToClient("<OPTION>Text<OPTION>HTML\012");
	&SendToClient("</SELECT><BR>\012");
	&SendToClient("<INPUT TYPE=\"submit\" VALUE=\"Post article\"><BR>Body of article:<BR>\012");
	&SendToClient("<TEXTAREA COLS=70 ROWS=12 NAME=\"Body\">$ReplyBody</TEXTAREA><BR>\012");
	&SendToClient("<INPUT TYPE=\"checkbox\" NAME=\"wraplines\" CHECKED>Wrap lines at column 70\012");
	&SendToClient("</FORM></BODY>\012");
}

# SendPostArticleResult($ProtoVers,$Newsgroups,$Subject,$Contenttype,$Body,$MessageIDRespondingTo)
# Posts the article and sends a page saying whether the post succeeded.

sub SendPostArticleResult {
	local($ProtoVers,$Newsgroups,$Subject,$Contenttype,$Body,$MessageIDRespondingTo) = @_;
	local($ArticleText) = "";
	
	if (!&ValidEmailAddress($Newsrc{"prefs","email"}) || $Newsrc{"prefs","realname"} eq "") {
		&Fail($ERR_NO_NAME_ENTERED,"You cannot post unless you enter your real name and e-mail address.");
	}
	
	$ArticleText = "From: ".$Newsrc{"prefs","email"}." (".$Newsrc{"prefs","realname"}.")\012";
	$ArticleText .= "Organization: ".$Newsrc{"prefs","organization"}."\012";

	# Deal with spaces in newsgroups line
	$Newsgroups =~ s/ /\,/g;			# turn spaces into commas
	$Newsgroups =~ s/\,[\,]+/,/g;		# combine multiple commas
	$Newsgroups =~ s/^\,//;				# remove any leading comma
	$Newsgroups =~ s/\,$//;				# remove any trailing comma
	
	$ArticleText .= "Newsgroups: $Newsgroups\012";
	$ArticleText .= "Subject: $Subject\012";
	if ($MessageIDRespondingTo ne "") {
		$ArticleText .= "References: <$MessageIDRespondingTo>\012";
	}
	$ArticleText .= "Mime-version: 1.0\012";
	$ArticleText .= "Content-type: $Contenttype\012";
	$ArticleText .= "\012";
	$Body =~ s/\015\012/\012/g;
	$Body =~ s/\015/\012/g;
	$Body =~ s/(^|\012)\./$1\.\./g;
	$ArticleText .= $Body;
	
	# Append the text or HTML signature, if it's not empty.
	local($Sig);
	if ($Contenttype eq "text/html") {
		$Sig = $Newsrc{"prefs","htmlsig"};
	} else {
		$Sig = $Newsrc{"prefs","textsig"};
	}
	if ($Sig ne "") {
		$Sig =~ s/\001/\012/g;
		$ArticleText .= "\012\012".$Sig;
	}
	
	# Add the final . so the NNTP server will see the end of the article.
	if ($ArticleText !~ /\012$/) {
		$ArticleText .= "\012";
	}
	$ArticleText .= ".\012";
	
#	print STDERR "About to post article: here's what I'll send to the server:\n";
#	local($Stuff) = $ArticleText;
#	$Stuff =~ s/\012/\n/g;
#	print STDERR $Stuff;
	
	local($ResponseCode,$ResponseText) = &SendNNTPCommand("POST\012");
	if ($ResponseCode != 340) {
		print STDERR "News server wouldn't allow posting: $ResponseCode $ResponseText\n";
		&Fail($ResponseCode,$ResponseText);
	}
	($ResponseCode,$ResponseText) = &SendNNTPCommand($ArticleText);
	if ($ResponseCode != 240) {
		print STDERR "News server wouldn't accept posted article: $ResponseCode $ResponseText\n";
		&Fail($ResponseCode,$ResponseText);
	}
	
	&SendProtoVers($ProtoVers,0);
	&SendToClient("<HEAD><TITLE>Article posted</TITLE></HEAD>\012");
	&SendToClient("<BODY><H3>The article was posted successfully</H3>\012");
	&SendToClient("It may take a few minutes to show up.<HR>\012");
	&SendToClient("Back to <A HREF=\"/Newsreading/NewsGroupPages/$CurrentNNTPServer/$CurrentGroup\">\"$CurrentGroup\" newsgroup</A></BODY>\012");
}

# SendPrefsErrorPage($ProtoVers,$ErrorTitle,$ErrorMessage,$ReturnURL)
# A user made an entry mistake on an account creation or personal preferences page.
# Send out an error message page.

sub SendPrefsErrorPage {
	local($ProtoVers,$ErrorTitle,$ErrorMessage,$ReturnURL) = @_;
	
	&SendProtoVers($ProtoVers,1);
	
	&SendToClient(
		"<HEAD><TITLE>$ErrorTitle</TITLE></HEAD>\012",
		"<BODY><H2>$ErrorTitle</H2>\012",
		"<P>$ErrorMessage</P>\012",
		"Back to $ReturnURL\012",
		"</BODY>\012"
	);
}

# SendHTTPInternalErrorMessage($ProtoVers,$ExceptionCode,$ExceptionText)

sub SendHTTPInternalErrorMessage {
	local($ProtoVers,$ExceptionCode,$ExceptionText) = @_;

	if ($ProtoVers ne "") {
		&SendToClient("HTTP/1.0 500 Internal error\012");
		# Without the following, some browsers assume that the message is plain text.
		&SendToClient("Content-type: text/html\012\012");
	}
	&SendToClient("<HEAD><TITLE>500 Internal error</TITLE></HEAD>\012");
	&SendToClient("<BODY>\012");
	&SendToClient("<H2>Your request could not be satisfied</H2>\012");
	if ($ExceptionCode == $ERR_GENERIC) {
		&SendToClient("$ExceptionText");
	} elsif ($ExceptionCode > 0) {
		&SendToClient("Sorry, something went wrong. The news server responded \"$ExceptionCode $ExceptionText\"\012");
	} else {
		&SendToClient("It was internal error $ExceptionCode \"$ExceptionText\".\012");
	}
	&SendToClient("</BODY>\012");
}

# SendHTTPNotFoundMessage($Url,$ProtoVers)
# sends a "404 Not Found" message for the given Url using the specified HTTP protocol version

sub SendHTTPNotFoundMessage {
	local($Url,$ProtoVers) = @_;

	if ($ProtoVers ne "") {
		&SendToClient("HTTP/1.0 404 Not found\012");
		# Without the following, some browsers assume that the message is plain text.
		&SendToClient("Content-type: text/html\012\012");
	}
	&SendToClient(
		"<HEAD><TITLE>404 Not Found</TITLE></HEAD>\012",
		"<BODY>\012",
		"<H1>404 Not Found</H1>\012",
		"<P>The requested URL $Url was not found on this server.</P>\012",
		"</BODY>\012"
	);
}

# SendTerminationMessage($ProtoVers)
# Sends a message saying that the server has quit.
# Eventually will include an informative message if an error
# occurred?

sub SendTerminationMessage {
	local($ProtoVers) = @_;

	&SendProtoVers($ProtoVers,0);
	&SendToClient(
		"<HEAD><TITLE>",
		"Server is shutting down",
		"</TITLE></HEAD>\012",
		"<BODY>\012",
		"<H1>The server is shutting down, as you requested.</H1>",
		"</BODY>\012"
	);
}

# SendRedirectionMessage($ProtoVers,$NewURL)
# Sends a 302 Redirection message pointing to the specified $NewURL.

sub SendRedirectionMessage {
	local($ProtoVers,$NewURL) = @_;
	
	if ($ProtoVers ne "") {
		&SendToClient(
			"HTTP/1.0 302 Found\012",
			"URI: $NewURL\012",
			"Location: $NewURL\012\012"
		);
	} else {
		# Then this routine is pretty pointless.
	}
}

# SendAuthorizationRequiredMessage($ProtoVers,$Realm)
# Sends a message saying that Authorization is necessary to access the URL.

sub SendAuthorizationRequiredMessage {
	local($ProtoVers,$Realm) = @_;
	
	if ($ProtoVers ne "") {
		&SendToClient(
			"HTTP/1.0 401 Unauthorized\012",
			"Date: ",&TimeString(time),"\012",
			"Mime-version: 1.0\012",
			"Content-type: text/html\012",
			"WWW-Authenticate: Basic realm=\"$Realm\"\012",
			"\012"
		);
	}
		
	&SendToClient(
		"<HEAD><TITLE>Authorization Required</TITLE></HEAD>\012",
		"<BODY><H1>Authorization Required</H1>\012",
		"Authentication failed. Did you type your password right? ",
		"Can your browser do authentication at all?</BODY>\012"
	);
}

1;
